import rumd, time
from pythonOutputManager import pythonOutputManager


class rumdSimulation:

    """
    Contains the main objects and provides basic simulation main loop.
    Once created at minimum the potential and integrator must be set
    before Run() can be called.

    """
    
    def __init__(self, filename, pb=32, tp=4, verbose=True, allow_tp_280_large=False):
        """
        Create the simulation object, and the underlying sample object, after
        reading in a configuration from the named file. Assign default values
        to various simulation- and output-related parameters. Parameters pb
        and tp set the number of particles per block and threads per particle
        respectively. verbose=False suppresses messages to standard output.
        allow_tp_280_large=True allows simulating large systems on GTX 280 cards
        with tp >1, which is known to cause problems.

        """
        self.sample = rumd.Sample(pb,tp, allow_tp_280_large)
        self.sample.SetVerbose(verbose)
        self.sample.ReadConf(filename)
               
        self.thermostat = False
        self.itg = None
        self.potentialList = []
        self.output_is_initialized = False
        self.blockSize = None
        self.moleculeData = None
        self.verbose = verbose
        self.report_time = False
        self.GPU_times = None
        
        self.outputParameters = {}
        self.outputParameters["energies"] = ("linear", 256)
        self.outputParameters["trajectory"] =  ("logarithmic",1)
        self.allowedManagers = ["energies","trajectory"]
        self.pyOutputList = []
        self.outputDirectory = "TrajectoryFiles" # at the moment this is specified separately in the sample constructor--potential for confusion

        self.momentum_reset_interval = 100
        self.sort_interval = 200
        self.exclusionInfo = {"Bond":set(),
                              "Angle":False,
                              "Dihedral":False}

    def SetTimingStatus(self, report_time):
        """
        Sets whether calls to Sample::CalcF should be timed on the GPU.
        If True, then after Run() the member self.GPU_times contains a list
        of pairs [total_force_time, nblist_update_time]. Used by the Autotuner
        (not necessary for the user to call).
        """
        self.report_time = report_time


    def SetVerbose(self, vb):
        """
        Turn on/off most messages in rumdSimulation and sample objects.
        sim.SetVerbose automatically calls sim.sample.SetVerbose; call
        to the latter explicitly to only affect sample messages.
        """
        self.verbose = vb
        self.sample.SetVerbose(self.verbose)

    def SetPB_TP(self, pb, tp):
        """
        Change pb and tp for the sample. Used mainly by the autotuner.
        Involves copying of data due array re-allocation.
        """
        nDOFs = self.sample.GetNumberOfDOFs()
        self.sample.SetPB_TP(pb,tp)        
        if self.moleculeData is not None:
            self.RecreateExclusionList()

        self.sample.SetNumberOfDOFs(nDOFs)

    def WritePotentials(self):
        """
        Write a file containing the pair potential
        and pair force for each associated pair potential.
        """
        for potential in self.potentialList:
            potential.WritePotentials(self.sample.GetSimulationBox())
            

    def WriteConstraints(self):
        """
        Write a file listing the deviations of bond lengths from
        constrained values.
        """
        if self.moleculeData is None:
            raise ValueError, "No MoleculeData object present"
        self.moleculeData.WriteConstraints()


    def ReadMoleculeData(self, topologyFile):
        """
        Read data describing how particles are connected as molecules
        and create the MoleculeData object.
        """
        self.moleculeData = rumd.MoleculeData(self.sample, topologyFile)
        self.sample.SetMoleculeData(self.moleculeData)
        
    def SetExclusionBond(self, type):
        """
        Exclude bonds of a given type. Not necessary to be called by user.
        """
        if len(self.potentialList) == 0:
            print "Warning: SetExclusionBond called before non-bonding potential has been set. Cannot set exclusion list."
        else:
            if self.verbose:
                print "Setting exclusion bond, type=",type
            for potential in self.potentialList:
                self.moleculeData.SetExclusionBond(potential, type)
                
    def SetBondHarmonic(self, type, lbond, ks, exclude=True):
        """
        Set a bond-type to be harmonic.
        """
        if self.moleculeData is None:
            raise ValueError, "No moleculeData object; call ReadMoleculeData first"
        self.moleculeData.SetBondHarmonic(type, lbond, ks)
        if exclude:
            self.exclusionInfo["Bond"].add(type) 
            self.SetExclusionBond(type)
        else:
            self.exclusionInfo["Bond"].discard(type) 
            
          
    def SetBondFENE(self, type, max_l, k):
        """
        Set a bond-type to be the FENE type.
        """
        if self.moleculeData is None:
            raise ValueError, "No moleculeData object; call ReadMoleculeData first"
        self.moleculeData.SetBondFENE(type, max_l, k)

    def SetExclusionAngle(self):
        """
        Exclude pair interactions associated with bond-angle interactions.
        Not necessary to be called by user.
        """
        if len(self.potentialList) == 0:
            print "Warning: SetExclusionAngle called before non-bonding potential has been set. Cannot set exclusion list"
        else:
            for potential in self.potentialList:
                self.moleculeData.SetExclusionAngle(potential)
                  
    def SetAngleCosSq(self, type, theta0, ktheta, exclude=True):
        """
        Set angle type coefficients.
        """
        if self.moleculeData is None:
            raise ValueError, "No moleculeData object; call ReadMoleculeData first"
        self.moleculeData.SetAngleCosSq(type, theta0, ktheta)

        if exclude:
            self.exclusionInfo["Angle"] = True 
            self.SetExclusionAngle()
        else:
            self.exclusionInfo["Angle"] = False
            
   
    def GetAngles(self):
        """
        Access array containing current values of bond angles.
        """
        if self.moleculeData is None:
            raise ValueError, "No moleculeData object; call ReadMoleculeData first"
        self.moleculeData.GetAngles()

    def GetDihedrals(self):
        """
        Access array containing current values of dihedral angles.
        """
        if self.moleculeData is None:
            raise ValueError, "No moleculeData object; call ReadMoleculeData first"
        self.moleculeData.GetDihedrals()

    def SetExclusionDihedral(self):
        """
        Exclude pair interactions associated with dihedral angle interactions. Not necessary to be called by user.
        """
        if len(self.potentialList) == 0:
            print "Warning: Set SetExclusionDihedral called before non-bonding potential has been set. Cannot set exclusion list"
        else:
            for potential in self.potentialList:
                self.moleculeData.SetExclusionDihedral(potential)
                
    def SetDihedralRyckaert(self, type, coeffs, exclude=True):
        """
        Set coefficients for a given type of dihedral angle, by default
        excluding corresponding pair interactions.
        """        
        if len(coeffs) != 6:
            raise ValueError, "The coefficient array must have length 6"
        
        if self.moleculeData is None:
            raise ValueError, "No moleculeData object; call ReadMoleculeData first"
        self.moleculeData.SetDihedralRyckaert(type, coeffs[0], coeffs[1],
                                              coeffs[2], coeffs[3],
                                              coeffs[4], coeffs[5])
        if exclude:
            self.exclusionInfo["Dihedral"] = True 
            self.SetExclusionDihedral()
        else:
            self.exclusionInfo["Dihedral"] = False

    def SetBondConstraint(self, type, lbond, exclude=True):
        """
        Set a bond-type as "constraint" and pass the length, by default
        excluding interactions from the pair potential for bonds of this type.
        """
        if self.moleculeData is None:
            raise ValueError, "No moleculeData object; call ReadMoleculeData first"
        
        self.moleculeData.SetBondConstraint(type, lbond)
        if exclude:
            self.exclusionInfo["Bond"].add(type) 
            self.SetExclusionBond(type)
        else:
            self.exclusionInfo["Bond"].discard(type) 

    def RecreateExclusionList(self):
        """
        Recreates the exclusion list via stored exclusion information. Not
        necessary to be called by user.
        """
        for type in self.exclusionInfo["Bond"]:
            self.SetExclusionBond(type)
        if self.exclusionInfo["Angle"]:
            self.SetExclusionAngle()
        if self.exclusionInfo["Dihedral"]:
            self.SetExclusionDihedral()

    def ScaleSystem(self, scaleFactor, dir=None):
        """
        Scale the simulation box and particle positions by scaleFactor
        in all directions, if dir is ommitted; otherwise only
        in the direction dir.
        """
        if dir is None:
            self.sample.IsotropicScaleSystem(scaleFactor)
        else:
            self.sample.AnisotropicScaleSystem(scaleFactor, dir)
        
    def SetSimulationBox(self, simBox):
        """
        Set an alternative simulationBox.
        """
        self.simulationBox = simBox # keep it alive by having a reference
        self.sample.SetSimulationBox(simBox)

    def SetPotential(self, potential):
        """
        Set a potential for this system, replacing any previous
        potentials.
        """
        self.potentialList = [potential]
        self.sample.SetPotential(potential)

    def AddPotential(self, potential):
        """
        Add a potential for this system, keeping any previous potentials
        (the total force vector and potential will be the sum over all
        added potentials).
        """
        self.potentialList.append(potential)
        self.sample.AddPotential(potential)

    def SetIntegrator(self, itg):
        """
        Set the integrator for this system, and call its momentum-resetting
        function.
        """
        self.sample.SetIntegrator(itg)
        self.itg = itg
        itg.SetMomentumToZero()
        
    def WriteConf(self, filename):
        """
        Write the current configuration to a file.
        """
        self.sample.WriteConf(filename)

    def SetBlockSize(self, blockSize):
        """
        Specify the block-size for output (None means automatic). Should be a
        power of 2 for logarithmic scheduling.
        """
        self.blockSize = blockSize

    def SetOutputDirectory(self, outputDirectory):
        """
        Change the directory for output files, default TrajectoryFiles.
        """
        self.outputDirectory = outputDirectory
        self.sample.SetOutputDirectory(outputDirectory)

    def GetNumberOfParticles(self):
        """
        Return the number of particles in the sample object.
        """
        return self.sample.GetNumberOfParticles()

    def GetVolume(self):
        """
        Return the volume defined by the simulation box.
        """
        return self.sample.GetSimulationBox().GetVolume()


    def SetOutputMetaData(self, manager, **kwargs):
        """
        Access to the output manager to control precision and what gets written.

        Examples
        ---------
        
        sim.SetOutputMetaData("trajectory", precision=6, virials=True)
        sim.SetOutputMetaData("energies", potentialEnergy=False)
        """
        if manager not in self.allowedManagers:
            raise ValueError, "SetOutputMetaData -- first argument must be one of" + str(self.allowedManagers)

        
        if manager in ["energies", "trajectory"]:
            metadataAccessObj = {"energies":self.sample.GetEnergiesOutputManager,
                              "trajectory":self.sample.GetTrajectoryConfMetaData}[manager]()
            for key in kwargs:
                if key == "precision":
                    metadataAccessObj.SetPrecision(key)
                else:
                    metadataAccessObj.Set(key, kwargs[key])
        else:
            raise ValueError, "Not implemented for manager %s yet" % manager


    def AddExternalCalculator(self, calc):
        """
        Add data computed by an external calculator class to the energies files.

        Example
        -------

        alt_pot_calc = AlternatePotentialCalculator(...)
        sim.AddExternalCalculator(alt_pot_calc)
        
        """
        self.sample.GetEnergiesOutputManager().AddExternalCalculator(calc)
        
    
    def NewOutputManager(self, name):
        """
        Create a new (python) output manager, which will write files in the
        output directory starting with name. 
        """
        self.pyOutputList.append( pythonOutputManager(name, self.outputDirectory) )
        self.allowedManagers.append(name)

    def SetOutputParameters(self, manager, schedule, **kwargs):
        """
        Old (deprecated) name for SetOutputScheduling.
        """
        print "[Info]: SetOutputParameters is deprecated. Use SetOutputScheduling instead"
        self.SetOutputScheduling(manager, schedule, **kwargs)

        
    def SetOutputScheduling(self, manager, schedule, **kwargs):
        """
        Set scheduling information for an output manager.

        manager must be one of the current managers, which include "energies"
        and "trajectory" and whatever other managers have been added.

        schedule must be one of "none", "linear","logarithmic","loglin"

        extra keyword arguments may/must be supplied where relevant, e.g.
        interval=100 for linear scheduling (required).
        base=10 for logarithmic scheduling (optional, default base is 1)
        base=1, maxInterval=16 for loglin (required)
        """
        allowedSchedules = ["none", "linear","logarithmic","loglin"]
        if manager not in self.allowedManagers:
            raise ValueError, "SetOutputScheduling -- first argument must be one of" + str(self.allowedManagers)
        if schedule not in allowedSchedules:
            raise ValueError, "SetOutputScheduling -- second argument must be " + str(allowedSchedules)

        if manager in ("energies","trajectory"):
            if schedule == "none":
                newParams = None
            elif schedule == "linear":
                newParams = ("linear",kwargs["interval"])
            elif schedule == "logarithmic":
                base = 1
                if kwargs.has_key("base"):
                    base = kwargs["base"]
                newParams = ("logarithmic",base)
            else:
                newParams = ("loglin",kwargs["base"],kwargs["maxInterval"])
            self.outputParameters[manager] = newParams

        else:
            for pyManager in self.pyOutputList:
                if pyManager.name == manager:
                    pyManager.SetOutputParameters(schedule, **kwargs)

    def RegisterCallback(self, manager, function, **kwargs ):
        """
        Register a data analysis function (which takes a Sample as argument)
        to be called with a specified time-step interval. Callback functions
        that are used to generate output should return a string without newline
        characters.
        """
        if manager not in self.allowedManagers[2:]:
            raise ValueError, "SetOutputParameters -- first argument must be one of" + str(self.allowedManagers[2:])
        for pyManager in self.pyOutputList:
            if pyManager.name == manager:
                pyManager.AddCallback(function, **kwargs)

    def SetMomentumResetInterval(self, mom_reset_interval):
        """
        Set how many time steps should go between resetting of
        center of mass momentum to zero.
        """
        self.momentum_reset_interval = mom_reset_interval

    def SetSortInterval(self, sort_interval):
        """
        Set how many time steps should go between sorting of
        particle positions (default value 200). Not necessary to be
        called by user if using autotune.
        """
        self.sort_interval = sort_interval

    def InitializeOutput(self, num_iter):
        """
        Initialize output managers at start of Run(),
        including choosing block size if not set by user.
        Not necessary to be called by the user.
        """
        SetLinearFuncMap = {"energies":self.sample.SetEnergyLinearWriteInterval,
                         "trajectory":self.sample.SetTrajectoryLinearWriteInterval}
        SetLogLinFuncMap = {"energies":self.sample.SetEnergyLogLinParams,
                            "trajectory":self.sample.SetTrajectoryLogLinParams}

        SetOnFuncMap = {"energies":self.sample.SetWriteEnergies,
                        "trajectory":self.sample.SetWriteTrajectory}

        ON = {}
        min_blockSize = 1;
        for manager in self.outputParameters.keys():
            outputParameters = self.outputParameters[manager]
            ON[manager] = (outputParameters is not None)
            SetOnFuncMap[manager](ON[manager])
            if ON[manager]:
                schedule = outputParameters[0]
                if schedule == "linear":
                    interval = outputParameters[1]
                    SetLinearFuncMap[manager](interval)
                    if interval > min_blockSize: min_blockSize = interval
                elif schedule == "logarithmic":
                    base = outputParameters[1]
                    SetLogLinFuncMap[manager](base, 0)
                else:
                    base = outputParameters[1]
                    maxInterval = outputParameters[2]
                    SetLogLinFuncMap[manager](base, maxInterval)

        if self.blockSize is None:
            nominalNumBlocks = 100
            if (ON["energies"] or ON["trajectory"]):
                nominalBlockSize = num_iter/nominalNumBlocks/2
                if min_blockSize < nominalBlockSize:
                    min_blockSize = nominalBlockSize
                blockSize = 1
                while blockSize < min_blockSize:
                    blockSize *= 2
            else:
                blockSize = num_iter/nominalNumBlocks
            
            self.blockSize = blockSize
        numBlocks = num_iter/self.blockSize + 1

        if self.verbose:
            print "Output blockSize=%d, numBlocks=%d" % (self.blockSize, numBlocks)
            print "Size of the last incomplete block", num_iter % self.blockSize
            
        # If blocksize is not set for python output managers it's the same.
        self.sample.SetOutputBlockSize(self.blockSize)
        for pyManager in self.pyOutputList:
            if pyManager.blockSize == None:
                pyManager.SetBlockSize(self.blockSize)

        self.output_is_initialized = True
   
    def Run(self, num_iter, initializeOutput=True,
            restartBlock=None, suppressAllOutput=False):
        """
        Run num_iter time steps, by default initializing/resetting output.
        Specify restartBlock to restart a simulation which has been itnerrupted.
        Set suppressAllOutput to True to avoid any output (for example for
        equilibration).
        """

        if self.itg is None:
            raise ValueError, "Integrator has not been set"
        if len(self.potentialList) == 0:
            raise ValueError, "Potential has not been set"


        initialTimeStep = 0
        if not suppressAllOutput:
            if initializeOutput:
                self.InitializeOutput(num_iter)
            else:
                if not self.output_is_initialized:
                    raise ValueError, "Cannot have initializeOutput=False if output has not previously been initialized"

            if restartBlock is not None:
                if not initializeOutput:
                    raise ValueError, "When restarting, initializeOutput must be True"
                self.sample.ReadRestartConf(restartBlock)
                initialTimeStep = restartBlock*self.blockSize

        # main loop
        GPU_times = []
        for i in xrange(initialTimeStep, num_iter):
            
            if self.momentum_reset_interval>0 and i%self.momentum_reset_interval == 0:
                self.sample.GetIntegrator().SetMomentumToZero()
            if self.sort_interval>0 and i%self.sort_interval == 0:
                self.sample.SortParticles()
                
            time_calcf = self.sample.CalcF(report_time=self.report_time) # Calculate forces
            if self.report_time:
                time_nb_update = 0.
                for potIdx in xrange(len(self.potentialList)):
                    time_nb_update += self.potentialList[potIdx].GetNbListTimeTaken()
                GPU_times.append([time_calcf, time_nb_update])

            if not suppressAllOutput:
                # Called here to ensure synchronization of positions and forces
                self.sample.NotifyOutputManagers(i)

            for manager in self.pyOutputList: # python managers for callback
                if manager.schedule == "none" or not suppressAllOutput:
                    if manager.nextCalcTimeStep == i:
                        manager.CalcOutput(self.sample)
                    if manager.nextWriteTimeStep == i:
                        manager.WriteOutput(self.sample)
            self.itg.Integrate() # Update positions and velocities
        # end main loop

        # empty last buffers with data
        for manager in self.pyOutputList:
            if manager.output != []:
                manager.WriteOutput(self.sample)

        if self.report_time:
            self.GPU_times = GPU_times
