#ifndef MOLECULEDATA_H
#define MOLECULEDATA_H

#include "rumd/Sample.h"
#include "rumd/SimulationBox.h"
#include "rumd/PairPotential.h"
#include "rumd/ConstraintPotential.h"
#include "rumd/BondPotential.h"
#include "rumd/AnglePotential.h"
#include "rumd/DihedralPotential.h"

#include <cuda.h>
#include <cfloat>

class ParticleData;

class MoleculeData {
 private:
  Sample* S;
  RectangularSimulationBox *SB;

  ConstraintPotential* constraintPotential;
  BondHarmonic *HarmBond;  
  BondFENE *FENEBond;
  AngleCosSq *AngleCos;
  DihedralRyckaert *DihedralRyc;
 
  unsigned num_mol;         // Number of molecules 
  unsigned max_num_uau;     // Maximum nuau per molecule

  unsigned max_num_bonds;    // Only used by constraints because this uses the old stuff.
  unsigned num_bonds;       // Maximum bonds per particle
  unsigned num_btypes;      // Number of bonding types 

  unsigned num_angles;      // Number of angles
  unsigned num_atypes;      // Number of angle types
  
  unsigned num_dihedrals;   // Number of dihedrals
  unsigned num_dtypes;      // Number of dihedral types

  std::map<std::string, unsigned int> allocatedValue;


  void AllocateMolecules();
  void AllocateBonds();
  void AllocateAngles();
  void AllocateDihedrals();

  void FreeMolecules();
  void FreeBonds();
  void FreeAngles();
  void FreeDihedrals();


 public:
  MoleculeData(Sample *set_sample, const std::string& top_filename);
  MoleculeData(const MoleculeData& M);
  ~MoleculeData();
  MoleculeData* Copy();
  MoleculeData& operator=(const MoleculeData& M);


  // Bond Lists/arrays
  std::vector<uint2> h_blist;  // Bonding list
                               // x, y: indices of bonded particles 
  uint2 *d_blist;
  std::vector<uint1> h_btlist; // Bond type list
  uint1  *d_btlist;
  float2 *h_bplist, *d_bplist; // Bond type parameter list (r0, k)
  float  *h_belist, *d_belist; // Potential energy for each bond
  int1   *h_btlist_int,        // Internal mapping from user specified type 
         *d_btlist_int;        //   to internal type

  // Angle lists/arrays
  std::vector<uint4> h_alist;  // Angle list:
  uint4  *d_alist;             // x, y, z: indices of bonded particles 
                               // w: Angle type
  float2 *h_aplist, *d_aplist; // Angle type parameter list 
                               // x: force constant
                               // y: zero force angle
  float  *h_epot_angle,        // Holds the potential energy for
         *d_epot_angle;        //   each angle

  // Dihedral lists/arrays
  std::vector<uint4> h_dlist; // Dihedral list:
  uint4 *d_dlist;                // x, y, z, w: indices of bonded particles
  std::vector<uint1> h_dtype; // Dihedral type list
  uint1 *d_dtype;
  float *h_dplist, *d_dplist; // Parameter lists            
  float *h_epot_dihedral,     // Holds the potential energy for
        *d_epot_dihedral;     //   each dihedral

  // Molecule lists (Convenient for data analysis)
  int1 *h_mlist;            
  int1 *d_mlist;            
  unsigned lmlist;          
  
  // Data arrays
  float *h_bonds, *d_bonds;              // The bonds 
  float *h_angles, *d_angles;            // The angles
  float *h_dihedrals, *d_dihedrals;      // The dihedrals

  float4 *h_cm, *d_cm;                   // Centre of mass + cm.w mass 
  float3 *h_vcm, *d_vcm;                 // Velocity centre of mass
  float3 *h_s, *d_s;                     // (Intrinsic) angular momentum
  float3 *h_omega, *d_omega;             // (Intrinsic) angular velocity
  float *h_inertia, *d_inertia;          // Moment of intertia (length=9*num_mols)
                                         // Is symmetric, but keep all element due to 
                                         // calculation of ang. velocity - row-wise matrix

  float *h_stress, *d_stress;            // Molecular stress tensor per molecule
  float symmetricStress[6];              // Molecule symmetric stress for the whole box

  // GPU specifics
  dim3 threads_per_block;  
  //dim3 num_blocks;

  // Helper functions 
  void AddmlistEntry(unsigned *, unsigned, unsigned, unsigned);
  void ReadTopology(const std::string& top_filename);
  void UpdateAfterSorting( unsigned int *d_new_index );
  

  // Set methods.
  void SetBondHarmonic(unsigned type, float lbond, float ks);
  void SetBondConstraint(unsigned type, float lbond);  
  void SetBondFENE(unsigned type, float max_lbond, float k);
  void SetAngleCosSq(unsigned type, float theta0, float ktheta);
  void SetDihedralRyckaert(unsigned type, float c0, float c1, float c2, float c3, 
			   float c4, float c5); // To incorp. in SWIG *sigh*
 
  void SetExclusionBond(PairPotential *potential, unsigned etype);
  void SetExclusionMolecule(PairPotential* potential);  
  void SetExclusionMolecule(PairPotential* potential, unsigned molindex);
  void SetExclusionAngle(PairPotential* potential);
  void SetExclusionDihedral(PairPotential* potential);
  void SetNumberOfConstraintIterations( unsigned numIte );
  void ResetConstraintLayout();

  void SetHarmonicBondConstant(unsigned type, float ks);
  void SetAngleConstant(unsigned type, float ktheta);
  void SetDihedralRyckaertConstants(unsigned type, float c0, float c1, float c2, 
				    float c3, float c4, float c5);

  // Get methods.
  unsigned GetNumberOfMolecules() const { return num_mol; };
  unsigned GetMaximumNumberOfBonds() const { return max_num_bonds; }
  unsigned GetNumberOfBonds() const { return num_bonds; }
  unsigned GetNumberOfAngles() const { return num_angles; }
  unsigned GetNumberOfDihedrals() const { return num_dihedrals; }

  const int1*  GetMoleculeListDevice() const { return d_mlist; }
  unsigned GetMaximumMoleculeSize() const { return max_num_uau; }

  void GetBondsFromTopology(const std::string& top_filename);
  void GetAnglesFromTopology(const std::string& top_filename);
  void GetDihedralsFromTopology(const std::string& top_filename);
  
  void GetBonds();
  void GetAngles();
  void GetDihedrals();
  
  // Write or output methods
  void WriteMolConf(const std::string& fstr);
  void WriteMolxyz(const std::string& fstr, unsigned mol_id, const std::string& mode);
  void WriteMolxyz_filtered(const std::string& fstr, unsigned filter_uau, const std::string& mode);
  void WriteConstraints();
  
  // Run time data evaluations
  void EvalMolMasses();
  void EvalCM(); 
  void EvalVelCM(); 
  void EvalSpin();
  //void EvalMolStressTensor(float cf);
  //void EvalStressACF(int lvec, float cf, float kBT, double isampleUnit, int opt);

  // Debugging.
  void PrintLists();
  void PrintLists(unsigned i);
  void PrintMlist();

};

// Kernel for updating Mlist after sorting
__global__ void UpdateMlistAfterSorting( unsigned int *new_index,
					 unsigned int nParticles,
					 int1 *mlist_offset,
					 unsigned length_mlist_offset );

__global__ void updateBondsAfterSorting(  unsigned int *new_index, 
					  unsigned int nParticles, 
					  uint2 *blist, 
					  unsigned num_bonds );

__global__ void updateAnglesAfterSorting( unsigned int *new_index, 
					  unsigned int nParticles, 
					  uint4 *alist, 
					  unsigned num_angles );

__global__ void updateDihedralsAfterSorting( unsigned int *new_index,
					     unsigned int nParticles,
					     uint4 *dlist,
					     unsigned num_dihedrals);

// Should of course be moved to a generic place
void GaussPivotBackCPU(float *x, float *b, float *A, unsigned nrc);
double Determinant3(float *A);

template <class Simbox>
__global__ void MolStressTensor(float *stress, float4 *r, float3 *vcm, float4 *rcm,
                                int1 *d_mlist, unsigned max_num_uau,
                                unsigned num_mol,  Simbox *simbox,
                                float *simboxpointer, float cf);

#endif // MOLECULEDATA_H
