#ifndef NEIGHBORLIST_H
#define NEIGHBORLIST_H

/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include "rumd/rumd_technical.h"
#include "rumd/rumd_base.h"
#include "rumd/ParticleData.h"
#include "rumd/KernelPlan.h"
#include <map>

class PairPotential; class SimulationBox;

class NeighborList{
 
 private:
  NeighborList(const NeighborList&); 
  NeighborList& operator=(const NeighborList& N);

  ParticleData* particleData;
  KernelPlan kp;
  SimulationBox* simBox;
  // Dynamic allocation
  unsigned int numBlocks;
  unsigned int numBlocksParticles;
  unsigned int numParticles;
  unsigned int particlesPerBlock;
  unsigned int numVirtualParticles; 
  unsigned int maxNumNbrBlocks;

  std::map<std::string, unsigned int> numberAllocatedUnits;

  int* d_updateNbList;
  float h_shell[MaxNumTypes*MaxNumTypes];
  float* d_shell;

  float skin;
  float max_cutoff;

  float* d_lastBoxShift; // for LE sim box, also includes deltaStrain and adjusted skin
  float2* d_dStrain;     // deltaStrain and 0.25 times the square of the strain-adjusted skin

  NBTYPEv* NbList;
  int* cell_list;
  float4* d_last_r;

  unsigned int* d_num_nbr_blocks;
  unsigned int* h_num_nbr_blocks; // temporary

  float4* d_Rc;          // Center of boundingbox for block
  float4* d_dR;          // Distance from surface to center in boundingbox
  
  unsigned* h_ExclusionType;
  NBTYPEv* h_ExclusionList;

  unsigned* d_ExclusionType;
  NBTYPEv* d_ExclusionList;
  NBTYPEv* d_OldExclusionList;
  NBTYPEv* d_savedNeighborList;


  float elapsedTimeLastUpdate;
  void AllocateLastPositions();
  void AllocateNumBlocks(); // arrays with length numBlocks
  void AllocateNeighborList();
  void AllocateSavedNeighborList();
  void AllocateExclusionType( unsigned int numberOfTypes );
  void AllocateExclusionList();
  unsigned int EstimateMaxNumNbrBlocks(SortingScheme ss);

 public:
  NeighborList();
  ~NeighborList();
  
  
  void Initialize( ParticleData* pData, KernelPlan& kPlan, SimulationBox* simBox);
  void UpdateNBlist(bool record_time_taken);
  void UpdateExclusionListAfterSorting( unsigned int* d_raw_ptr_new_index );
  void CopyExclusionListToDevice();
  void CopyExclusionTypeToDevice();
  void SaveNeighborList();
  void RestoreNeighborList();
  void ResetNeighborList();
  void CheckAllocatedNbrBlocks();
  void SetMaxNumNbrBlocks(unsigned max_num_nbr_blocks);

  // Output.
  void PrintExclusionList();
  void PrintCellList();
  void PrintBoundingboxes();

  // Set methods.
  void SetCutoffs(PairPotential* potential);
  void SetCutoffs(PairPotential* potential, float skin);
  void SetExclusionList( NBTYPEv* newList );
  
  void SetExclusion( unsigned particleI, unsigned particleJ );
  void SetExclusionType( unsigned type0, unsigned type1 );
  void SetExclusionBond(uint1* h_btlist, uint2* h_blist, unsigned num_part, unsigned num_virt_part,unsigned max_num_bonds, unsigned etype); 
  void SetExclusionAngle(uint4* h_alist, unsigned num_angles);
  void SetExclusionDihedral(uint4* h_dlist, unsigned num_dihedrals);
  void SetExclusionMolecule(int1* h_mlist, unsigned molindex, unsigned max_num_uau, unsigned num_mol);
  
  // Get methods.
  
  NBTYPEv* GetNbListPtr() const {return NbList;} // not const in deep sense
  int* GetCellListPtr() const { return cell_list; }
  unsigned int GetMaxNumNbrBlocks() const { return maxNumNbrBlocks; }

  float GetSkin() const { return skin; }
  float GetMaxCutoff() const { return max_cutoff; }
  float GetElapsedTimeLastUpdate() const { return elapsedTimeLastUpdate; }
  unsigned int GetMaxUsedNbrBlocks();
};


__global__ void update_deltaStrain(float* lastBoxShift, float2* dStrain, float* simBoxPointer, float skin, float max_cutoff);
__global__ void UpdateMyExclusionListAfterSorting( unsigned int* revIndex, NBTYPEv* oldExclusionList, NBTYPEv* exclusionList, unsigned numParticles );

template <class S> __global__ void displacement_since_nb_build( unsigned numParticles, float4* d_r, float4* d_last_r, float4* d_im,
								S* simBox, float* simBoxPointer, float skin2over_4, int* updateNbList, float2* dStrain);

template <class S> __global__ void calc_boundingbox( unsigned int num_part, float4* r, float4* Rc, float4* dR, 
						     S* simBox, float* simBoxPointer, float maxCutPlusSkin, int* updateNbList );

template <class S> __global__ void calc_cell_list( unsigned int num_blocks, unsigned int maxNumNbrBlocks, float4* Rc, float4* dR, 
						   S* simBox, float* simBoxPointer, int* cell_list, int* updateNbList, unsigned int* num_nbr_blocks );


template <class S> __global__ void md_update_nblist( unsigned int num_part, unsigned int maxNumNbrBlocks, float4* r, float4* im, float* params, 
						     S* simBox, float* simBoxPointer, NBTYPEv* NbList, NBTYPEv* ExclusionList, 
						     unsigned* d_ExclusionType, int* cell_list, int *updateNbList, float4* last_r, float* lastBoxShift, unsigned numberOfTypes );

#endif // NEIGHBORLIST_H
