
/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

#include "rumd/BondPotential.h"
#include "rumd/SimulationBox.h"
#include "rumd/MoleculeData.h"


__device__ inline void BondAtomicFloatAdd(float *address, float val){
  int i_val = __float_as_int(val);
  int tmp0 = 0;
  int tmp1;
  while( (tmp1 = atomicCAS((int *)address, tmp0, i_val)) != tmp0){
    tmp0 = tmp1;
    i_val = __float_as_int(val + __int_as_float(tmp1));
  }
}

__global__ void BondResetForces( float4 *f, float4 *w, float4* sts ){
  float4 zero = {0.0f, 0.0f, 0.0f, 0.0f};
  f[blockIdx.x] = zero;
  w[blockIdx.x] = zero;
  sts[blockIdx.x] = zero;
}


///////////////////////////////////////////////////////
// Harmonic potential
///////////////////////////////////////////////////////

BondHarmonic::BondHarmonic(MoleculeData *globalM){
  M = globalM;
  SetID_String("bondHarmonic");

  num_bonds = M->GetNumberOfBonds();
  num_threads = 32;
  num_blocks = num_bonds/num_threads + 1;
}


float BondHarmonic::CalcF(bool initialize, bool calc_stresses, bool report_time){

  cudaEvent_t start;
  cudaEvent_t stop;
  float elapsedTime = 0.;
  
  if(report_time) {
    cudaEventCreate( &start );
    cudaEventCreate( &stop );
    cudaEventRecord( start, 0);
  }

  if ( initialize )
    BondResetForces<<<particleData->GetNumberOfParticles(),1>>>(particleData->d_f, particleData->d_w, particleData->d_sts);
  
  RectangularSimulationBox* RSB = dynamic_cast<RectangularSimulationBox*>(simBox);
  LeesEdwardsSimulationBox* LESB = dynamic_cast<LeesEdwardsSimulationBox*>(simBox);

  if ( RSB )
    if( calc_stresses )
      Harmonic<2, 1><<<num_blocks,num_threads>>>( particleData->d_r, 
						  particleData->d_f,
						  particleData->d_w,
						  particleData->d_sts,
						  M->d_blist, M->d_btlist, 
						  M->d_bplist, M->d_belist, 
						  M->d_btlist_int, M->d_bonds,
						  num_bonds, RSB, 
						  RSB->GetDevicePointer() );
    else
      Harmonic<1, 1><<<num_blocks,num_threads>>>( particleData->d_r,
						  particleData->d_f,
						  particleData->d_w,
						  particleData->d_sts,
						  M->d_blist, M->d_btlist, 
						  M->d_bplist, M->d_belist, 
						  M->d_btlist_int, M->d_bonds,
						  num_bonds, RSB, 
						  RSB->GetDevicePointer() );


  else if ( LESB )
    if ( calc_stresses )
      Harmonic<2, 1><<<num_blocks,num_threads>>>( particleData->d_r,
						  particleData->d_f,
						  particleData->d_w,
						  particleData->d_sts,
						  M->d_blist, M->d_btlist,
						  M->d_bplist, M->d_belist, 
						  M->d_btlist_int, M->d_bonds,
						  num_bonds, LESB,
						  LESB->GetDevicePointer());
    else
      Harmonic<1, 1><<<num_blocks,num_threads>>>( particleData->d_r,
						  particleData->d_f,
						  particleData->d_w,
						  particleData->d_sts,
						  M->d_blist, M->d_btlist,
						  M->d_bplist, M->d_belist, 
						  M->d_btlist_int, M->d_bonds,
						  num_bonds, LESB,
						  LESB->GetDevicePointer());
  else
    throw RUMD_Error("BondHarmonic","CalcF","unknown simulation box");
 
  if(report_time) {
    cudaEventRecord( stop, 0 );
    cudaEventSynchronize( stop );
    cudaEventElapsedTime( &elapsedTime, start, stop);
    cudaEventDestroy(start);
    cudaEventDestroy(stop);
  }
  
  return elapsedTime;
}

double BondHarmonic::GetPotentialEnergy(){
   
  RectangularSimulationBox* RSB = dynamic_cast<RectangularSimulationBox*>(simBox);
  LeesEdwardsSimulationBox* LESB = dynamic_cast<LeesEdwardsSimulationBox*>(simBox);

  if ( RSB )
    Harmonic< 1, 2 ><<<num_blocks, num_threads>>>
      ( particleData->d_r, particleData->d_f,
	particleData->d_w, particleData->d_sts,
	M->d_blist, M->d_btlist, M->d_bplist,
	M->d_belist, M->d_btlist_int, M->d_bonds,
	num_bonds, RSB, RSB->GetDevicePointer() );
  else if ( LESB )
    Harmonic<1, 2><<<num_blocks, num_threads>>>
      ( particleData->d_r, particleData->d_f,
	particleData->d_w, particleData->d_sts,
        M->d_blist, M->d_btlist, M->d_bplist,
        M->d_belist, M->d_btlist_int, M->d_bonds,
        num_bonds, LESB, LESB->GetDevicePointer() );
  else
    throw RUMD_Error("BondHarmonic","GetPotentialEnergy","unknown simulation box");

  // copy to host
  size_t nbytes = num_bonds*sizeof(float);
  cudaMemcpy(M->h_belist, M->d_belist, nbytes, cudaMemcpyDeviceToHost);
  
  // sum over bonds
  double pe = 0.0;
  for ( unsigned i=0; i<num_bonds; i++ ){
    unsigned bond_type = M->h_btlist[i].x;
    if ( M->h_btlist[bond_type].x == 0 )
      pe += (double)M->h_belist[i];
  }  

  return pe;
}

///////////////////////////////////////////////////////
// FENE potential
///////////////////////////////////////////////////////

BondFENE::BondFENE(MoleculeData *globalM ){

  M = globalM;
  SetID_String("bondFene");

  num_bonds = M->GetNumberOfBonds();
  num_threads = 32;
  num_blocks = num_bonds/num_threads + 1;
}

float BondFENE::CalcF(bool initialize, bool calc_stresses, bool report_time){

  cudaEvent_t start;
  cudaEvent_t stop;
  float elapsedTime = 0.;

  if(report_time) {
    cudaEventCreate( &start );
    cudaEventCreate( &stop );
    cudaEventRecord( start, 0);
  }

  if ( initialize )
    BondResetForces<<<particleData->GetNumberOfParticles(),1>>>( particleData->d_f, particleData->d_w, particleData->d_sts );

  RectangularSimulationBox* RSB = dynamic_cast<RectangularSimulationBox*>(simBox);
  LeesEdwardsSimulationBox* LESB = dynamic_cast<LeesEdwardsSimulationBox*>(simBox);

  if ( RSB )
    if( calc_stresses )
      FENE<2, 1><<<num_blocks, num_threads>>>
	( particleData->d_r, particleData->d_f,
	  particleData->d_w, particleData->d_sts,
	  M->d_blist, M->d_btlist, M->d_bplist,
	  M->d_belist, M->d_btlist_int, M->d_bonds,
	  num_bonds, RSB, RSB->GetDevicePointer() );
    else
      FENE<1, 1><<<num_blocks, num_threads>>>
	( particleData->d_r, particleData->d_f,
	  particleData->d_w, particleData->d_sts,
	  M->d_blist, M->d_btlist, M->d_bplist,
	  M->d_belist, M->d_btlist_int, M->d_bonds,
	  num_bonds, RSB, RSB->GetDevicePointer() );
  else if ( LESB )
    if( calc_stresses )
      FENE<2, 1><<<num_blocks, num_threads>>>
	( particleData->d_r, particleData->d_f,
	  particleData->d_w, particleData->d_sts,
	  M->d_blist, M->d_btlist, M->d_bplist,
	  M->d_belist, M->d_btlist_int, M->d_bonds,
	  num_bonds, LESB, LESB->GetDevicePointer() );
    else
      FENE<1, 1><<<num_blocks, num_threads>>>
	( particleData->d_r, particleData->d_f,
	  particleData->d_w, particleData->d_sts,
	  M->d_blist, M->d_btlist, M->d_bplist,
	  M->d_belist, M->d_btlist_int, M->d_bonds,
	  num_bonds, LESB, LESB->GetDevicePointer() );
  else
    throw RUMD_Error("BondFENE","CalcF","unknown simulation box");

  if(report_time) {
    cudaEventRecord( stop, 0 );
    cudaEventSynchronize( stop );
    cudaEventElapsedTime( &elapsedTime, start, stop);
    cudaEventDestroy(start);
    cudaEventDestroy(stop);
  }

  return elapsedTime;
}

double BondFENE::GetPotentialEnergy(){
  
  RectangularSimulationBox* RSB = dynamic_cast<RectangularSimulationBox*>(simBox);
  LeesEdwardsSimulationBox* LESB = dynamic_cast<LeesEdwardsSimulationBox*>(simBox);
  if ( RSB )
    FENE< 1, 2 ><<<num_blocks, num_threads>>>
      ( particleData->d_r, particleData->d_f,
	particleData->d_w, particleData->d_sts,
	M->d_blist, M->d_btlist, M->d_bplist,
	M->d_belist, M->d_btlist_int, M->d_bonds,
	num_bonds, RSB, RSB->GetDevicePointer() );
  else if ( LESB )
    FENE<1, 2><<<num_blocks, num_threads>>>
      ( particleData->d_r, particleData->d_f,
	particleData->d_w, particleData->d_sts,
	M->d_blist, M->d_btlist, M->d_bplist,
	M->d_belist, M->d_btlist_int, M->d_bonds,
	num_bonds, LESB, LESB->GetDevicePointer() );
  else
    throw RUMD_Error("BondFENE","GetPotentialEnergy","unknown simulation box");
  
  // copy to host
  size_t nbytes = num_bonds*sizeof(float);
  cudaMemcpy(M->h_belist, M->d_belist, nbytes, cudaMemcpyDeviceToHost);
  
  // sum over (relevant) bonds
  double pe = 0.0;
  for ( unsigned i=0; i<num_bonds; i++ ){
    unsigned bond_type = M->h_btlist[i].x;
    if ( M->h_btlist_int[bond_type].x == 1 )
      pe +=  (double)M->h_belist[i];
  }  

  return pe;
}

///////////////////////////////////////////////////////
// Kernel implementations of bond potentials
///////////////////////////////////////////////////////

template <int stress, int energy, class Simbox> 
__global__ void Harmonic( float4 *r, float4 *f, float4 *w, float4 *sts,
			  uint2 *blist, uint1 *btlist, float2 *bplist,
			  float *belist, int1 *btlist_int, float *bonds,
			  unsigned num_bonds, Simbox *simbox, float *simboxpointer ){

  unsigned bond_index = blockIdx.x*blockDim.x + threadIdx.x;

  if ( bond_index < num_bonds ){

    unsigned bond_type = btlist[bond_index].x;

    if ( btlist_int[bond_type].x == 0 ){

      // Get bond information
      unsigned a   =  blist[bond_index].x;
      unsigned b   =  blist[bond_index].y;
      float r0 = bplist[bond_type].x; // bond length
      float k  = bplist[bond_type].y; // spring constant
      
      // Load the simulation box in local memory to avoid bank conflicts.
      float array[simulationBoxSize];

      simbox->loadMemory(array, simboxpointer);
      
      // Calculate stuff
      float4 dr = simbox->calculateDistance(r[a], r[b], array);
      float rij = sqrt(dr.w); // squared distance
      bonds[bond_index] = rij;

      float pe = 0.5f*k*( rij-r0 )*( rij-r0 ); // potential energy

      if ( energy == 2 )
	belist[bond_index] = pe;
      else if ( energy == 1 ){
	float ft = -k*(1.0f - r0/rij); // force
	BondAtomicFloatAdd(&(f[a].x), ft*dr.x);
	BondAtomicFloatAdd(&(f[a].y), ft*dr.y);
	BondAtomicFloatAdd(&(f[a].z), ft*dr.z);
	BondAtomicFloatAdd(&(f[a].w), 0.5f*pe);
	
	BondAtomicFloatAdd(&(f[b].x), -ft*dr.x);
	BondAtomicFloatAdd(&(f[b].y), -ft*dr.y);
	BondAtomicFloatAdd(&(f[b].z), -ft*dr.z);
	BondAtomicFloatAdd(&(f[b].w), 0.5f*pe);
	
	if ( stress >= 1 ){
	  float virial = ft*dr.w; // *( dr.x*dr.x + dr.y*dr.y + dr.z+dr.z );
	  BondAtomicFloatAdd(&(w[a].w), virial ); 
	  BondAtomicFloatAdd(&(w[b].w), virial ); 
	  if( stress == 2 ) {
	    float my_stress[6] = {-ft*dr.x*dr.x, -ft*dr.y*dr.y,
				  -ft*dr.z*dr.z, -ft*dr.y*dr.z,
				  -ft*dr.x*dr.z, -ft*dr.x*dr.y};
	    
	    BondAtomicFloatAdd(&(sts[a].x), my_stress[0] );
	    BondAtomicFloatAdd(&(sts[a].y), my_stress[1] );
	    BondAtomicFloatAdd(&(sts[a].z), my_stress[2] );
	    BondAtomicFloatAdd(&(sts[a].w), my_stress[3] );
	    BondAtomicFloatAdd(&(w[a].y), my_stress[4] );
	    BondAtomicFloatAdd(&(w[a].z), my_stress[5] );
	    
	    BondAtomicFloatAdd(&(sts[b].x), my_stress[0] );
	    BondAtomicFloatAdd(&(sts[b].y), my_stress[1] );
	    BondAtomicFloatAdd(&(sts[b].z), my_stress[2] );
	    BondAtomicFloatAdd(&(sts[b].w), my_stress[3] );
	    BondAtomicFloatAdd(&(w[b].y), my_stress[4] );
	    BondAtomicFloatAdd(&(w[b].z), my_stress[5] );
	  }
	  
	}
      }
    }
  }
}


template <int stress, int energy, class Simbox>
__global__ void FENE( float4 *r, float4 *f, float4 *w, float4 *sts,
		      uint2 *blist, uint1 *btlist, float2 *bplist,
		      float *belist, int1 *btlist_int, float *bonds, 
		      unsigned num_bonds, Simbox *simbox, float *simboxpointer ){
  
  unsigned bond_index = blockIdx.x*blockDim.x + threadIdx.x;

  if( bond_index < num_bonds ){

    unsigned bond_type = btlist[bond_index].x;

    if( btlist_int[bond_type].x == 1 ){

      // Get bond information
      unsigned a = blist[bond_index].x; // particle index
      unsigned b = blist[bond_index].y; // particle index
      float r0 = bplist[bond_type].x; // max bond length
      float k0 = bplist[bond_type].y; // prefactor

      // Load the simulation box in local memory to avoid bank conflicts.
      float array[simulationBoxSize];
      simbox->loadMemory(array, simboxpointer);

      // Do calculation
      float4 dr = simbox->calculateDistance(r[a], r[b], array);

      float rij = sqrt(dr.w); // not necesary for force 
      bonds[bond_index] = rij;

      float rr = dr.w / (r0*r0);
      float ft = - k0/(1.0f - rr);
      float pe = - 0.5f* r0*r0 * k0 * logf(1.0f - rr);

      if ( energy == 2 )
	belist[bond_index] = pe;
      else if ( energy == 1 ){
	BondAtomicFloatAdd(&(f[a].x), ft*dr.x);
	BondAtomicFloatAdd(&(f[a].y), ft*dr.y);
	BondAtomicFloatAdd(&(f[a].z), ft*dr.z);
	BondAtomicFloatAdd(&(f[a].w), 0.5f*pe);

	BondAtomicFloatAdd(&(f[b].x), -ft*dr.x);
	BondAtomicFloatAdd(&(f[b].y), -ft*dr.y);
	BondAtomicFloatAdd(&(f[b].z), -ft*dr.z);
	BondAtomicFloatAdd(&(f[b].w), 0.5f*pe);
	if ( stress >= 1 ){
	  float virial = ft*dr.w;// *( dr.x*dr.x + dr.y*dr.y + dr.z+dr.z );
	  BondAtomicFloatAdd(&(w[a].w), virial );
	  BondAtomicFloatAdd(&(w[b].w), virial );
	  
	  if( stress == 2 ) {
	    float my_stress[6] = {-ft*dr.x*dr.x, -ft*dr.y*dr.y,
				  -ft*dr.z*dr.z, -ft*dr.y*dr.z,
				  -ft*dr.x*dr.z, -ft*dr.x*dr.y};
	  
	    BondAtomicFloatAdd(&(sts[a].x), my_stress[0] );
	    BondAtomicFloatAdd(&(sts[a].y), my_stress[1] );
	    BondAtomicFloatAdd(&(sts[a].z), my_stress[2] );
	    BondAtomicFloatAdd(&(sts[a].w), my_stress[3] );
	    BondAtomicFloatAdd(&(w[a].y), my_stress[4] );
	    BondAtomicFloatAdd(&(w[a].z), my_stress[5] );
	    
	    BondAtomicFloatAdd(&(sts[b].x), my_stress[0] );
	    BondAtomicFloatAdd(&(sts[b].y), my_stress[1] );
	    BondAtomicFloatAdd(&(sts[b].z), my_stress[2] );
	    BondAtomicFloatAdd(&(sts[b].w), my_stress[3] );
	    BondAtomicFloatAdd(&(w[b].y), my_stress[4] );
	    BondAtomicFloatAdd(&(w[b].z), my_stress[5] );
	  }


	}
      }
    }
  }
}


